import numpy
import matplotlib

matplotlib.use("agg")
from matplotlib import pyplot

import pytest

import lsys
from lsys import viz

BASELINE_DIR = "baseline_images"
TOLERANCE = 15

plotData = numpy.array(
    [
        3.113,
        3.606,
        4.046,
        4.046,
        4.710,
        6.140,
        6.978,
        2.000,
        4.200,
        4.620,
        5.570,
        5.660,
        5.860,
        6.650,
        6.780,
        6.790,
        7.500,
        7.500,
        7.500,
        8.630,
        8.710,
        8.990,
        9.850,
        10.820,
        11.250,
        11.250,
        12.200,
        14.920,
        16.770,
        17.810,
        19.160,
        19.190,
        19.640,
        20.180,
        22.970,
    ]
)


@pytest.mark.parametrize(
    ("xlim", "ylim", "expected"),
    [
        ([6, 12], [-12, 10], ([-2, 20], [-12, 10])),
        ([30, 60], [-5, 10], ([30, 60], [-12.5, 17.5])),
    ],
)
def test_square_aspect(xlim, ylim, expected):
    result = viz.square_aspect(xlim, ylim)
    assert result == expected


@pytest.mark.mpl_image_compare(
    baseline_dir=BASELINE_DIR,
    tolerance=TOLERANCE,
)
def test_square_aspect2():
    numpy.random.seed(42)
    y = plotData
    x = numpy.random.uniform(-1, 1, size=len(y)) * 5

    fig, ax = pyplot.subplots(figsize=(4, 4))
    _ = viz.plot(x, y, ax=ax, color="b", marker="o", linestyle="")
    viz.pretty_format_ax(ax=ax, x=x, y=y)

    return fig


@pytest.mark.parametrize(
    "seq",
    [
        ["red", "green", "blue"],
        ["red", "green", "blue", "blue"],
        ["red", "green", 0.33, "green", "blue", 0.66, "blue"],
        ["r", "g", "b", "b"],
        ["red", "green"],
    ],
)
@pytest.mark.mpl_image_compare(
    baseline_dir=BASELINE_DIR,
    tolerance=TOLERANCE,
    remove_text=True,
)
def test_make_colormap_plot(seq):
    numpy.random.seed(42)
    cmap = viz.make_colormap(seq)
    fig, ax = pyplot.subplots(figsize=(3, 3), dpi=300)
    N = 1000
    array_dg = numpy.random.uniform(0, 10, size=(N, 2))
    x, y = array_dg[:, 0], array_dg[:, 1]
    colors = array_dg[:, 1]
    cax = ax.scatter(x, y, c=colors, cmap=cmap, lw=0)
    cbar = fig.colorbar(cax)
    return fig


@pytest.mark.parametrize(
    "fractal",
    [
        ("Bush1"),
        ("Bush2"),
        ("Crosses"),
        ("Dragon"),
        ("Dragon45"),
        ("Gosper"),
        ("Hexdragon"),
        ("Hilbert"),
        ("Penrose_Snowflake"),
        ("Plant_a"),
        ("Plant_b"),
        ("Plant_c"),
        ("Plant_d"),
        ("Plant_e"),
        ("Plant_f"),
        ("Putmans_Tattoo"),
        ("QuadKochIsland"),
        ("Serpinski_Curve"),
        ("Serpinski_Gasket"),
        ("SquareSpikes"),
        ("Terdragon"),
        ("Tree1"),
        ("Tree2"),
        ("Tree3"),
        ("Twig"),
        ("Two_Ys"),
        ("Weed1"),
        ("Weed2"),
        ("Weed3"),
    ],
)
@pytest.mark.mpl_image_compare(
    baseline_dir=BASELINE_DIR,
    tolerance=TOLERANCE,
    remove_text=True,
    savefig_kwargs={"dpi": 150},
)
def test_plot_lsys(fractal):
    kwargs = lsys.fractals.Fractal[fractal]
    f = lsys.Lsys(**kwargs)
    f.unoise = 0
    fig, axes = pyplot.subplots(1, 4, figsize=(12, 3))
    depths = [0, 1, 2, 4]
    for ax, depth in zip(axes, depths):
        f.depth = depth
        ax = f.plot(ax=ax, color="k", alpha=0.5, square=True)
        ax = f.plot(ax=ax, as_lc=True, color="k", alpha=0.5, square=True)
        ax.set_title(fractal)
    pyplot.close("all")
    return fig


@pytest.mark.parametrize(
    "fractal",
    [
        ("Dragon"),
        ("Terdragon"),
        ("Serpinski_Gasket"),
        ("Tree1"),
        ("SquareSpikes"),
        ("Plant_f"),
    ],
)
@pytest.mark.mpl_image_compare(
    baseline_dir=BASELINE_DIR,
    tolerance=TOLERANCE,
    remove_text=True,
    savefig_kwargs={"dpi": 150},
)
def test_plot_bezier_lsys(fractal):
    kwargs = lsys.fractals.Fractal[fractal]
    f = lsys.Lsys(**kwargs)
    f.unoise = 0
    fig, axes = pyplot.subplots(1, 4, figsize=(12, 3))
    depths = [0, 1, 2, 4]
    for ax, depth in zip(axes, depths):
        f.depth = depth
        ax = f.plot_bezier(ax=ax, color="k", alpha=0.33, square=True)
        ax = f.plot_bezier(ax=ax, as_lc=True, color="k", alpha=0.33, square=True)
        ax = f.plot_bezier(
            ax=ax,
            as_lc=True,
            segs=10,
            color="k",
            alpha=0.33,
            square=True,
        )

        ax.set_title(fractal)
    pyplot.close("all")
    return fig



"""
This module has functions helpful in the construction of general
bezier curves, as well as fitting cubic bezier curves between
vertices calculated by the :class:``lsys.Lsys``

In addition, this module contains the code used to determine the
symetric weight factor needed to determine the position of control
points that form near-circular curves between three points with
an arbitrary interior angle, buiding on the work documented
`here <http://spencermortensen.com/articles/bezier-circle/>`_.
"""

from __future__ import division
import numpy

from . import validate
from . import algo


def ctrl_pts(p0, p1, w):
    """
    Calculate the position of a new point along vector
    `p0` to `p1` with magnitude `w`.

    Parameters
    ----------
    p0 : 1D array
        the starting point
    p1 : 1D array
        the end point
    w : float
        the weight factor

    Returns
    -------
    point : 1D array
        a point along vector `p0` to `p1` with magnitude `w`
    """
    p0 = validate.is_np(p0)
    p1 = validate.is_np(p1)
    vec = p1 - p0
    return p0 + vec * w


def binomial(n, k):
    """
    Parameters
    ----------
    n : integer
        the order of the binomial to expand
    k : integer
        the zero indexed position of the binomial expansion
        coefficient to return

    Returns
    -------
    coefficient : integer
        coefficient for the nth order binomial expansion, kth term

    reference: http://pomax.github.io/bezierinfo/
    """

    tri = [
        [1],  # n=0
        [1, 1],  # n=1
        [1, 2, 1],  # n=2
        [1, 3, 3, 1],  # n=3
        [1, 4, 6, 4, 1],  # n=4
        [1, 5, 10, 10, 5, 1],  # n=5
        [1, 6, 15, 20, 15, 6, 1],  # n=6
    ]

    while n >= len(tri):
        s = len(tri)
        nextrow = [0 for i in range(s + 1)]
        nextrow[0] = 1
        prev = s - 1
        for i in range(1, s):
            nextrow[i] = tri[prev][i - 1] + tri[prev][i]
        nextrow[s] = 1
        tri.append(nextrow)
    return tri[n][k]


def poly(n, k):
    """
    Expands binomial and computes value at `t`

    Parameters
    ----------
    n : integer
        nth order of polynomial
    k : integer
        positional term for binomial expansion coefficient

    Returns
    -------
    value : float or list of floats


    Reference
    ---------
    https://gist.github.com/Juanlu001/7284462
    """
    coeff = binomial(n, k)

    def _poly(t):
        """Calculates value at t between bezier points"""
        return coeff * (1 - t) ** (n - k) * t ** (k)

    return _poly


def bezier(points, segs=100):
    """
    Computes nth order bezier curves through the `points` provided
    and returns the curve as a list of points to produce line segments.

    The order of the bezier curve is determined by the length of
    the list of points. The length of the output curve numpy.array is
    equal to `segs` + 1, such that `segs` is equal to the number of
    line segments between the points of the array.

    Parameters
    ----------
    points : list-like
        2D array
    segs : integer, optional (default = 100)
        The number of line segments to approximate bezier curve

    Returns
    -------
    curve : numpy.array
        2D array of points along the curve.

    Reference
    ---------
    https://gist.github.com/Juanlu001/7284462
    """

    points = validate.is_np(points)
    dims = points.shape[1]

    # increment num such that final line segments will be equal to `segs`
    num = segs + 1

    n = len(points)
    t = numpy.linspace(0, 1, num=num)
    curve = numpy.zeros((num, dims), dtype=float)
    for i, pt in enumerate(points):
        curve += numpy.outer(poly(n - 1, i)(t), pt)

    return curve


def circular_weight(angle):
    """This function utilizes the precomputed circular bezier function
    with a fit to a 10th order curve created by the following code block:

    .. code-block:: python

        x = numpy.arange(.5, 180, 0.5)
        y = []
        for i in x:
            y.append(bezier.find_circular_weight(i, tol=1e-12, max_iters=500))
        y = numpy.array(y)
        z = numpy.polyfit(x, y, 10)


    Parameters
    ----------
    angle : float
        enter the angle to be traversed in degrees

    Returns
    -------
    weight : float
        Weight value for calculating bezier control points that
        approximate a circular curve.

    """
    z = numpy.array(
        [
            -2.45143082907626980583458614241573e-24,
            1.58856196152315352138612607918623e-21,
            -5.03264989277462933391916020538014e-19,
            8.57954915199159887348249578203777e-17,
            -1.09982713519619074150585319501519e-14,
            6.42175701661701683377126465867012e-13,
            -1.95012445981222027957307425487916e-10,
            6.98338125134285339870680633234242e-10,
            -1.27018636324842636571531492850617e-05,
            5.58069196465371404519196542326487e-08,
            6.66666581437823202449521886592265e-01,
        ]
    )

    p = numpy.poly1d(z)

    return p(angle)


def find_circular_weight(angle, guess=0.5, tol=1e-9, max_iters=50, r=10, segs=100):
    r"""Finds weight factor for bezier control points such that the `angle`
    QRS will be filled with cubic bezier curve approximating a circle
    that is tangent at the points Q and S.

    Parameters
    ----------
    angle : float
        the turn angle in degrees. e.g., if turn angle is 30 degrees, then the angle
        that needs to be spanned by the curve is 150 degrees.
    guess : float, optional (default=0.5)
        the initial value for the bisection search
    tol : float, optional (default=1e-9)
        the tolerance for ceasing bisection search
    max_iters : int, optional (default=50)
        the maximum iterations for bisection search
    r : float, optional (default=10)
        the radius for the circle. the larger this value, the more precise the
        weight will be since tolerance is absolute.
    segs : int, optional (defualt=100)
        number of points to consider on the circle, and bezier points to compute.

    Returns
    -------
    weight : float
        the magnitude for scaling control points relative to start and end points.
    error_np : 1D numpy.array
        an array of float error values from the curve to circular radius
    error : float
        the max error minus min error
    iterations : integer
        the count of bisection search iterations

    References
    ----------
    http://spencermortensen.com/articles/bezier-circle/


               ^
               | /
               |/   \
    h -- -- -- R    angle
              / \   /
             / | \
            /  |  \
           /  o o  \
          /o   |   o\
         /o    |    o\
    0 - Q o----+----o S -- >
               0

    """

    if not 0 < angle < 180:
        raise ValueError("`angle` must be between 0 and 180, exclusive")

    if not 0 < guess < 1:
        raise ValueError("`guess` must be between 0 and 1, exclusive")

    deg = (180 - angle) / 2
    theta = numpy.deg2rad(deg)  # half of the interior angle formed by QRS

    r = r
    # y location of acute angle QRS apex
    h = r * numpy.cos(theta) / numpy.tan(theta)
    x = 0  # x location of acute angle QRS apex
    kc = -h * numpy.tan(theta) * numpy.tan(theta)  # y value of circle center
    xc = h * numpy.tan(theta)  # x value of circle and line RS tangent point

    pt0 = numpy.array([-xc, 0])  # position of point Q
    pt2 = numpy.array([xc, 0])  # position of point S
    ptm = numpy.array([x, h])  # position of apex point R

    a = 1  # upper bound for bisection search
    b = 0  # lower bound for bisection search

    guess_gen = gen_new_guess(guess, pt0, pt2, ptm, kc, r, a, b, segs)
    for i in range(max_iters):

        guess, error_np, error, a, b = next(guess_gen)
        if -tol < error < tol:
            return guess, error_np, error, i

    return guess, error_np, error, i  # type: ignore


def gen_new_guess(guess, pt0, pt2, ptm, kc, r, a, b, segs):
    """Generator to produce the next guess in the bisection search sequence"""
    while True:
        pc1 = ctrl_pts(pt0, ptm, guess)  # first bezier control point along QR
        pc2 = ctrl_pts(pt2, ptm, guess)  # second bezier control point along RS
        # build cubic bezier curve with `segs` segments
        t = bezier([pt0, pc1, pc2, pt2], segs)
        # determine radial distance from each bezier point to the circle center
        radial_dist = numpy.sqrt(numpy.sum((t - [0, kc]) ** 2, axis=1))
        # radial distance from circle to bezier curve
        error_np = radial_dist - r
        tmax = numpy.abs(error_np.max())
        tmin = numpy.abs(error_np.min())
        error = tmax - tmin  # minimize the range with bisection search

        yield guess, error_np, error, a, b

        if error > 0:
            a = guess
        else:
            b = guess
        guess = (a + b) / 2


def bezier_xy(x, y, weight=None, angle=90, segs=100, keep_ends=True):
    """
    Compute cubic bezier curves between each set of three points
    using symetric control points offset by the `weight`.

    Parameters
    ----------
    x : list-like
        a 1D array of x values
    y : list-like
        a 1D array of y values
    weight : float, optional (default = None)
        Weight given to intermediate bezier control points. If
        `None` then the weight will be determined by the `angle`
    angle : float, optional (default = 90)
        the interior angle for the bezier sections to traverse
    segs : integer, optional (default = 100)
        the number of segments for bezier curve approzimation between
        each set of three points
    keep_ends : bool, optional (default = True)
        include start and end coordinates in output

    Returns
    -------
    tx : numpy.array
        an array of x values
    ty : numpy.array
        an array of y values

    """

    if weight is None:
        weight = circular_weight(angle)

    _x = [x[s] for s in numpy.ma.clump_unmasked(numpy.ma.masked_invalid(x))]
    _y = [y[s] for s in numpy.ma.clump_unmasked(numpy.ma.masked_invalid(y))]

    tx = []
    ty = []

    insert_nan = False
    if len(_x) > 1:
        insert_nan = True

    for segx, segy in zip(_x, _y):

        xmid, ymid = algo.midpoints(segx), algo.midpoints(segy)
        temp = numpy.vstack((xmid, ymid)).T
        rng = numpy.arange(0, len(temp) - 1, 2)
        _tx = []
        _ty = []
        last_pt = None
        for i in rng:
            pt = temp[i : i + 3]
            # unsure what this is doing here.
            if i > 0 and not numpy.array_equal(pt[0], last_pt):  # type: ignore # pragma: no cover
                continue
            c1 = ctrl_pts(pt[0], pt[1], weight)
            c2 = ctrl_pts(pt[2], pt[1], weight)
            t = bezier([pt[0], c1, c2, pt[2]], segs)
            _tx.append(t[:, 0])
            _ty.append(t[:, 1])
            last_pt = pt[2]

        flat_tx = numpy.array(_tx).flatten()
        flat_ty = numpy.array(_ty).flatten()

        if keep_ends:
            flat_tx = numpy.concatenate((segx[:1], flat_tx, segx[-1:]))
            flat_ty = numpy.concatenate((segy[:1], flat_ty, segy[-1:]))

        tx.append(flat_tx)
        ty.append(flat_ty)

        if insert_nan:
            for t in [tx, ty]:
                t.append([numpy.nan])

    tx = numpy.concatenate(tx)
    ty = numpy.concatenate(ty)

    return tx, ty



import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure

import pytest

from lsys import validate


def _pyplot():

    from matplotlib import pyplot

    return pyplot


def test_axes_object_invalid():
    with pytest.raises(ValueError):
        validate.axes_object("junk")


def test_axes_object_with_ax():
    fig, ax = _pyplot().subplots()
    fig1, ax1 = validate.axes_object(ax)
    assert isinstance(ax1, Axes)
    assert isinstance(fig1, Figure)
    assert ax1 is ax
    assert fig1 is fig


def test_axes_object_with_None():
    fig1, ax1 = validate.axes_object(None)
    assert isinstance(ax1, Axes)
    assert isinstance(fig1, Figure)


@pytest.mark.parametrize(
    ("listlike", "expected"),
    [
        ([0.5, 0.5], np.array([0.5, 0.5], dtype=float)),
        (np.array([0.5, 0.5], dtype=float), np.array([0.5, 0.5], dtype=float)),
    ],
)
def test_is_np(listlike, expected):
    result = validate.is_np(listlike)
    np.array_equal(result, expected)



import numpy
import matplotlib
from matplotlib.collections import LineCollection, PathCollection
from matplotlib.path import Path

from . import algo
from . import bezier
from . import validate


def plot(x, y, ax=None, **kwargs):
    """Plots 'x' vs 'y' using matplotlib. This is a convenience wrapper that
    creates the axes if it doesn't exist, and sets some nice rendering defaults
    for plotting l-systems.

    Parameters
    ----------
    x : numpy.array
        1D numpy array of points
    y : numpy.array
        1D numpy array of points
    ax : matplotlib.axes.Axes, optional (default = None)
        the axes artist for plotting. If `None` is given, a new figure
        will be created.
    kwargs : dict, optional
        all remaining keyword arguments are passed into the `plot()` command

    Returns
    -------
    ax : matplotlib.axes.Axes

    """
    if "solid_capstyle" not in kwargs:  # pragma: no cover
        kwargs["solid_capstyle"] = "round"

    _, axes = validate.axes_object(ax)
    _ = axes.plot(x, y, **kwargs)
    return axes


def plot_collection(collection, ax=None):
    _, axes = validate.axes_object(ax)
    axes.add_collection(collection)
    return axes


def construct_line_collection(coords, **kwargs):
    if "cmap" in kwargs and "array" not in kwargs:  # pragma: no cover
        kwargs["array"] = numpy.linspace(0.0, 1.0, len(coords))

    if "capstyle" not in kwargs:  # pragma: no cover
        kwargs["capstyle"] = "round"

    lines = LineCollection(coords, **kwargs)

    return lines


def bezier_segment_mpl(x, y, angle=None, weight=None, keep_ends=None):

    if angle is not None and weight is None:  # pragma: no cover
        weight = bezier.circular_weight(angle)
    if weight is None:  # pragma: no cover
        raise ValueError("one of angle or weight are required.")
    xmid, ymid = algo.midpoints(x), algo.midpoints(y)
    temp = numpy.vstack((xmid, ymid)).T
    rng = numpy.arange(0, len(temp) - 1, 2)
    paths = []
    for i in rng:
        pt = temp[i : i + 3]
        a = pt[0]
        b = bezier.ctrl_pts(pt[0], pt[1], weight)
        c = bezier.ctrl_pts(pt[2], pt[1], weight)
        d = pt[2]
        verts = [a, b, c, d]
        codes = [1, 4, 4, 4]  # {1: Path.MOVETO, 4: Path.CURVE4}
        paths.append(Path(verts, codes))
    if keep_ends:
        pi = [Path([[x[0], y[0]], [xmid[0], ymid[0]]])]
        pe = [Path([[xmid[-1], ymid[-1]], [x[-1], y[-1]]])]
        return pi + paths + pe
    return paths


def bezier_paths_mpl(x, y, weight=None, angle=None, keep_ends=None):

    if angle is not None and weight is None:  # pragma: no cover
        weight = bezier.circular_weight(angle)
    if weight is None:  # pragma: no cover
        raise ValueError("one of angle or weight are required.")

    _x = [x[s] for s in numpy.ma.clump_unmasked(numpy.ma.masked_invalid(x))]
    _y = [y[s] for s in numpy.ma.clump_unmasked(numpy.ma.masked_invalid(y))]

    paths = []
    for segx, segy in zip(_x, _y):
        _paths = bezier_segment_mpl(segx, segy, weight=weight, keep_ends=keep_ends)
        paths.extend(_paths)
    return paths


def construct_bezier_path_collection(coords, **kwargs):

    angle = kwargs.pop("angle", None)
    weight = kwargs.pop("weight", None)
    keep_ends = kwargs.pop("keep_ends", None)

    if angle is not None and weight is None:  # pragma: no cover
        weight = bezier.circular_weight(angle)
    if weight is None:  # pragma: no cover
        raise ValueError("one of angle or weight are required.")

    x, y = algo.coords_to_xy(coords)

    paths = bezier_paths_mpl(x, y, weight=weight, angle=angle, keep_ends=keep_ends)

    if "cmap" in kwargs and "array" not in kwargs:  # pragma: no cover
        kwargs["array"] = numpy.linspace(0.0, 1.0, len(paths))
    if "capstyle" not in kwargs:  # pragma: no cover
        kwargs["capstyle"] = "round"
    if "facecolor" not in kwargs:  # pragma: no cover
        kwargs["facecolor"] = "none"

    pc = PathCollection(paths, **kwargs)

    return pc


def pretty_format_ax(ax, x=None, y=None, coords=None, pad=5, square=None):
    """
    Parameters
    ----------
    ax : matplotlib.axes.Axes
        the axes artist for plotting.
    x : numpy.array, optional
        1D numpy array of points
    y : numpy.array, optional
        1D numpy array of points
    coords : numpy.array, optional
        2D numpy array of points
    pad : float or int, optional (default = 5)
        the percent spacing between plot data and axes lines. This can be used
        to zoom the plot.
    square : bool, optional (default = False)
        whether to make the output plot a square with equal aspect
        ratios for the axes.

    Returns
    -------
    ax : matplotlib.axes.Axes

    """

    if square is None:
        square = True

    if coords is not None:
        xlim, ylim = get_coord_lims(coords, pad=pad, square=square)
    else:
        xlim, ylim = get_xy_lims(x, y, pad=pad, square=square)

    _ = ax.set_xlim(xlim)
    _ = ax.set_ylim(ylim)
    _ = ax.set_xticks([])
    _ = ax.set_yticks([])

    if square:  # pragma: no branch
        _ = ax.set_aspect("equal")
    ax.get_figure().subplots_adjust(
        left=0.05, right=1 - 0.05, bottom=0.05, top=1 - 0.05
    )
    return ax


def square_aspect(xlim, ylim):
    """Calculates the limits to produce a square plot if all axes are
    equal scale.

    """
    x0, x1 = xlim
    y0, y1 = ylim

    x_range = numpy.abs(x1 - x0)
    y_range = numpy.abs(y1 - y0)

    if x_range > y_range:
        fac = (x_range - y_range) / 2

        return xlim, [y0 - fac, y1 + fac]

    else:
        fac = (y_range - x_range) / 2

        return [x0 - fac, x1 + fac], ylim


def pad_lim(lim, pad=None):
    if pad is None:
        pad = 0  # pragma: no cover
    _min, _max = lim
    pad = (_max - _min) * pad / 100
    lim = [_min - pad, _max + pad]

    return lim


def get_coord_lims(coords, pad=None, square=None):

    x, y = algo.coords_to_xy(coords)

    return get_xy_lims(x, y, pad=pad, square=square)


def get_xy_lims(x, y, pad=None, square=None):

    x0, x1 = numpy.nanmin(x), numpy.nanmax(x)
    y0, y1 = numpy.nanmin(y), numpy.nanmax(y)
    xlim, ylim = pad_lim([x0, x1], pad=pad), pad_lim([y0, y1], pad=pad)

    if square:
        return square_aspect(xlim, ylim)

    return xlim, ylim


def make_colormap(seq):
    """Return a LinearSegmentedColormap with colors defined as the sequence.

    The sequence red to blue can be defined as:
    `make_colormap(['red', 'blue'])`
    or equivalently
    `make_colormap([(1,0,0), 'blue'])`

    For more control over color transitions, floats can be included after pairs
    of colors to indicate the percentage of the color map that the transition
    should take place.

    `make_colormap(['red', 'green', .25, 'green', 'blue'])`




    Parameters
    ----------
    seq : list
        a sequence of floats, strings, or RGB-tuples.
        The floats should be in increasing order
        and in the interval (0, 1).

    Returns
    -------
    cmap : matplotlib.colors.LinearSegmentedColormap

    Examples
    --------

    References
    ----------
    http://stackoverflow.com/questions/16834861/create-own-colormap-using-matplotlib-and-plot-color-scale
    """
    CC = matplotlib.colors.ColorConverter()
    LSC = matplotlib.colors.LinearSegmentedColormap

    temp1 = seq
    if all(isinstance(x, str) for x in seq):
        if len(seq) > 2:
            temp1 = []
            f = 1.0 / (len(seq) - 1)
            for i in range(len(seq) - 1):
                temp1.append(seq[i])
                temp1.append(seq[i + 1])
                if i < len(seq) - 2:
                    temp1.append(f * (i + 1))

    temp2 = []
    for color in temp1:
        if isinstance(color, str):
            color = CC.to_rgb(color)
        temp2.append(color)
    seq = [(None,) * 3, 0.0] + temp2 + [1.0, (None,) * 3]
    cdict = {"red": [], "green": [], "blue": []}
    for i, item in enumerate(seq):
        if isinstance(item, float):
            r1, g1, b1 = seq[i - 1]
            r2, g2, b2 = seq[i + 1]
            cdict["red"].append([item, r1, r2])
            cdict["green"].append([item, g1, g2])
            cdict["blue"].append([item, b1, b2])
    return LSC("CustomMap", cdict)



import numpy

from . import validate


def add_noise(num, scale=0.5):
    """produces a random value distributed normally between +- num"""
    if num:
        return num * numpy.random.normal(0.0, scale)
    return 0


def midpoints(np_array):
    """
    Converts np array to an array of midpoints ignoring

    Parameters
    ----------
    np_array : array-like
        1-dimensional data

    Returns
    -------
    1-D numpy.array
        The length of the output is one less than
        the length of the input array.

    """
    i = 1

    np_array = validate.is_np(np_array)

    result = (
        np_array[~numpy.isnan(np_array)][i:] + np_array[~numpy.isnan(np_array)][:-i]
    ) / 2

    return result


def coords_to_xy(coords):
    """
    Parameters
    ---------
    coords : numpy.array
        a 2D array

    Returns
    -------
    x : numpy.array
        a 1D array
    y :
    """
    x = coords[:, :, 0].flatten()
    y = coords[:, :, 1].flatten()

    return x, y


def xy_to_coords(x, y, stride=None):
    _x = validate.is_np(x)
    _y = validate.is_np(y)

    segments = numpy.array([_x, _y]).T.reshape(-1, 1, 2)
    segments = numpy.concatenate([segments[:-1], segments[1:]], axis=1)

    if stride is not None:
        return segments[0::stride]

    return segments



from matplotlib.axes import Axes
import numpy


def axes_object(ax):
    """Checks if a value if an Axes. If None, a new one is created.
    Both the figure and axes are returned (in that order).

    """

    if ax is None:
        from matplotlib import pyplot

        fig, ax = pyplot.subplots()

    elif isinstance(ax, Axes):
        fig = ax.figure
    else:
        msg = "`ax` must be a matplotlib Axes instance or None"
        raise ValueError(msg)

    return fig, ax


def is_np(listlike):
    if not isinstance(listlike, numpy.ndarray) or listlike.dtype != float:
        listlike = numpy.array(listlike, dtype=float)

    return listlike



import warnings
import numpy

from . import viz
from . import bezier
from . import algo

MAX_STRING_SIZE = 5e6


class Lsys(object):
    """
    This object combines the basic language and directives
    necessary to construct a Lindenmayer system, compute
    the coordinates based on the directives, and plot the
    resulting figure.

    Attributes
    ----------
    string :
    depth
    coord
    x
    y

    References
    ----------


    """

    def __init__(
        self,
        axiom=None,
        rule=None,
        depth=0,
        a0=90,
        da=90,
        step=1,
        ds=1,
        unoise=0,
        forward="F",
        bar="|",
        right="+",
        left="-",
        goto="G",
        ignore="",
        name="",
        memory_check=True,
    ):
        """
        Parameters
        ----------




        """

        # settable properties
        self._axiom = None
        if axiom is not None:
            self._axiom = str(axiom).upper()

        self._rule = None
        if rule is not None:
            self._rule = rule
            self._clean_rule = self.clean_rule(rule)

        self._depth = depth
        self._a0 = a0
        self._da = da
        self._step = step
        self._ds = ds
        self._unoise = unoise
        self._forward = forward.upper()
        self._bar = bar.upper()
        self._left = left.upper()
        self._right = right.upper()
        self._goto = goto.upper()
        self._ignore = ignore.upper()
        self._name = name
        if not isinstance(memory_check, bool):  # pragma: no cover
            raise ValueError("`memory_check` must be `True` or `False`.")
        self._memory_check = memory_check

        # derived properties
        self._vocab = None
        self._commands = None
        self._coords = None
        self._depths = None
        self._x = None
        self._y = None
        self._bezier_coords = None
        self._bezier_x = None
        self._bezier_y = None
        self._string = None
        self._string_stale = True
        self._coord_stale = True
        self._bezier_stale = True

    # TODO: create __repr__ method
    def __repr__(self):
        rep = "<Lsys({})>"
        args = [
            "axiom",
            "rule",
            "depth",
            "a0",
            "da",
            "step",
            "ds",
            "unoise",
            "forward",
            "bar",
            "right",
            "left",
            "goto",
            "ignore",
            "name",
            "memory_check",
        ]
        reprs = []
        for arg in args:

            attr = getattr(self, arg)
            if isinstance(attr, str):
                val = "'" + attr + "'"
            else:
                val = attr
            reprs.append("{}={}".format(arg, val))

        return rep.format(", ".join(reprs))

    # settable properties
    @property
    def axiom(self):
        return self._axiom

    @axiom.setter
    def axiom(self, value):
        self._axiom = str(value).upper()
        self._string_stale = True
        self._coord_stale = True

    @property
    def rule(self):
        return self._rule

    @rule.setter
    def rule(self, value):
        self._rule = value
        self._clean_rule = self.clean_rule(value)
        self._string_stale = True
        self._coord_stale = True

    @property
    def depth(self):
        return self._depth

    @depth.setter
    def depth(self, value: int):
        self._depth = value
        self._string_stale = True
        self._coord_stale = True

    @property
    def a0(self):
        return self._a0

    @a0.setter
    def a0(self, value):
        self._a0 = value
        self._coord_stale = True

    @property
    def da(self):
        return self._da

    @da.setter
    def da(self, value):
        self._da = value
        self._coord_stale = True

    @property
    def step(self):
        return self._step

    @step.setter
    def step(self, value):
        self._step = value
        self._coord_stale = True

    @property
    def ds(self):
        return self._ds

    @ds.setter
    def ds(self, value):
        self._ds = value
        self._coord_stale = True

    @property
    def unoise(self):
        return self._unoise

    @unoise.setter
    def unoise(self, value):
        self._unoise = value
        self._coord_stale = True

    @property
    def forward(self):
        return self._forward

    @forward.setter
    def forward(self, value):
        self._forward = value.upper()
        self._coord_stale = True

    @property
    def bar(self):
        return self._bar

    @bar.setter
    def bar(self, value):
        self._bar = value.upper()
        self._coord_stale = True

    @property
    def left(self):
        return self._left

    @left.setter
    def left(self, value):
        self._left = value.upper()
        self._coord_stale = True

    @property
    def right(self):
        return self._right

    @right.setter
    def right(self, value):
        self._right = value.upper()
        self._coord_stale = True

    @property
    def goto(self):
        return self._goto

    @goto.setter
    def goto(self, value):
        self._goto = value.upper()
        self._coord_stale = True

    @property
    def ignore(self):
        return self._ignore

    @ignore.setter
    def ignore(self, value):
        self._ignore = value.upper()
        self._coord_stale = True

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def memory_check(self):
        return self._memory_check

    @memory_check.setter
    def memory_check(self, value):
        if not isinstance(value, bool):  # pragma: no cover
            raise ValueError("Must set `memory_check` to `True` or `False`.")
        self._memory_check = value

    # derived properties
    @property
    def string(self):
        if not self._string or self._string_stale:
            self._string = self.expand(
                self.axiom,
                self._clean_rule,
                self.depth,
                self.bar,
                self.memory_check,
            )
            self._string_stale = False
        return self._string

    @property
    def vocab(self):
        if self._vocab is None:
            self._vocab = self._build_vocab()
        return self._vocab

    @property
    def commands(self):
        if self._commands is None:
            self._commands = self._build_commands()
        return self._commands

    @property
    def depths(self):
        if self._coords is None or self._coord_stale:
            self._coords, self._depths = self.compute_coords()
            self._x, self._y = None, None
            self._coord_stale = False
        return self._depths

    @property
    def coords(self):
        if self._coords is None or self._coord_stale:
            self._coords, self._depths = self.compute_coords()
            self._x, self._y = None, None
            self._coord_stale = False
        return self._coords

    @property
    def x(self):
        if self._x is None or self._coord_stale:
            self._x, self._y = algo.coords_to_xy(self.coords)
        return self._x

    @property
    def y(self):
        if self._y is None or self._coord_stale:
            self._x, self._y = algo.coords_to_xy(self.coords)
        return self._y

    def _build_vocab(self):
        """Compile all chars used in the vocabulary of the fractal"""
        vocab = ""
        for k, v in self._clean_rule.items():
            vocab += k
            vocab += v
        vocab += "".join(
            [
                str(self.axiom),
                self.forward,
                self.goto,
                self.right,
                self.left,
                self.bar,
                self.ignore,
            ]
        )
        vocab = set(vocab.replace(" ", ""))

        if "." in vocab:
            raise Exception('the "." charcter is reserved and cannot be in `vocab`.')

        return vocab

    def _build_commands(self):
        """Compile all chars used in the vocabulary of the fractal"""
        cmd = ""
        cmd += "".join([self.forward, self.goto, self.right, self.left, self.bar])

        return set(cmd.replace(" ", ""))

    @staticmethod
    def clean_rule(rule):
        """
        Clean and validate the rule mapping.

        The defaults allow syntax conventions seen in the
        "Algorithmic Botany", "Computational Beauty of Nature", and
        "Matters Computational: Ideas, Algorithms, Source Code"

        Parameters
        ----------
        rule : string or dict
            a mapping for expanding characters of the string.
            If a string is used, the format requires a valid
            separator between key and value. These include ":",
            "=", "-->", "->", "=>". If multiple rules are provided,
            they must be separated by ";" or ",".


        Returns
        -------
        rule : dict
        """
        if isinstance(rule, dict):
            clean_rule = {}
            for k, v in rule.items():
                clean_rule[k.strip().upper()] = v.strip().upper()

            return clean_rule

        elif isinstance(rule, str):
            rule = rule.upper().replace(" ", "")

            seps = [":", "=", "-->", "->", "=>"]
            divs = [";", ","]
            clean_rule = {}

            if any(s in rule for s in seps):
                for s in seps:
                    if s in rule:
                        rule = rule.replace(s, "=")
            else:
                raise ValueError(
                    "Invalid rule syntax, use {} " "to assign rules.".format(seps)
                )
            for d in divs:
                if d in rule:
                    rule = rule.replace(d, ",")

            rules_tmp = rule.split(",")

            for r in rules_tmp:
                k, v = r.split("=")
                clean_rule[k.strip()] = v.strip()

            return clean_rule

        else:
            raise ValueError("`rule` must be string or mapping")

    @staticmethod
    def expand(axiom, rule, depth, bar="|", memory_check=True):
        """
        String expansion implemting `rule` on `axiom` for `depth` iterations.

        Parameters
        ----------
        axiom : string
            the starting string
        rule : dict
            the rules for expanding the characters of the string
        depth : integer
            the number of times to perform the expansion
        bar : string, optional (default = "|")
            the character to indicate that a calculation should be done
            by the parser based on the present depth of the string expansion
        memory_check : bool, optional (default = True)
            stop expanding when things get too large

        Returns:
        string : str
            the string expansion for the given depth


        """

        i = 0
        while i <= depth:
            if memory_check and len(axiom) > MAX_STRING_SIZE:
                raise MemoryError("Maximum `depth` is: {}".format(i))
            if i == 0:
                output = axiom
            else:
                output = ""
                for c in axiom:
                    if c in rule:
                        output += rule[c]
                    else:
                        output += c
            # axiom = output.replace(bar, str(i + 1) + '.')
            # this broke tests. check CBN to see if the ds values are given.
            axiom = output.replace(bar, str(i) + ".")
            i += 1
        return axiom.replace(".", bar)

    def _compute_bezier(self, bezier_weight=None, segs=None, keep_ends=None):
        """compute new bezier curves"""

        if segs is None:
            segs = 50
        if keep_ends is None:
            keep_ends = True

        self._bezier_x, self._bezier_y = bezier.bezier_xy(
            self.x,
            self.y,
            angle=self.da,
            weight=bezier_weight,
            segs=segs,
            keep_ends=keep_ends,
        )
        self._bezier_coords = algo.xy_to_coords(self._bezier_x, self._bezier_y)

        return self._bezier_x, self._bezier_y

    def compute_coords(self):
        """
        Parse the string and convert to data coordinates for rendering.

        Parameters
        ----------
        self : lsys.Lsys()
            uses most of the attributes from the class constructor
            to parse the expanded string and turn it into coordinates
            for a plotter.

        Returns
        -------
        x_y : numpy.array

        depths : numpy.array

        """
        a = numpy.radians(self.a0)
        da = numpy.radians(self.da)
        depth = self.depth
        unoise = self.unoise
        step = self.step
        ds = self.ds
        string = self.string
        forward = self.forward
        bar = self.bar
        left = self.left
        right = self.right
        goto = self.goto
        ignore = self.ignore

        tol = 1e-9

        x = 0
        y = 0
        sx = 0
        sy = 0

        x_y = []
        stack = []
        depths = []
        num = 0
        pres_depth = depth
        found = False
        for c in string:
            if c.isdigit():
                num = num * 10 + int(c)

            else:
                if c in (forward + bar + goto):

                    if c == bar:
                        pres_depth = num
                    else:
                        pres_depth = depth
                    s = step * ds**pres_depth * (algo.add_noise(unoise) + 1)

                    sy = y + numpy.sin(a + algo.add_noise(unoise)) * s
                    sx = x + numpy.cos(a + algo.add_noise(unoise)) * s

                    if c == goto:
                        x_y.append(([numpy.nan, numpy.nan], [numpy.nan, numpy.nan]))
                        depths.append(pres_depth)
                        pass

                    else:
                        ax = x if numpy.abs(x) > tol else 0
                        ay = y if numpy.abs(y) > tol else 0
                        bx = sx if numpy.abs(sx) > tol else 0
                        by = sy if numpy.abs(sy) > tol else 0
                        x_y.append(([ax, ay], [bx, by]))
                        depths.append(pres_depth)

                    x = sx
                    y = sy

                    found = True

                if c in (right + left):
                    num = num or 1
                    if c == right:
                        a -= da * num * (algo.add_noise(unoise) + 1)
                    else:
                        a += da * num * (algo.add_noise(unoise) + 1)

                elif c == "[":

                    stack.append(
                        (x, y, a + (algo.add_noise(unoise) * numpy.radians(5)))
                    )

                elif c == "]":
                    x, y, a = stack.pop()

                    x_y.append(([numpy.nan, numpy.nan], [numpy.nan, numpy.nan]))

                    depths.append(pres_depth)

                elif c in ignore:
                    pass

                elif not found:
                    ignore += c
                    warnings.warn("The {} character was ignored".format(c))

                num = 0
            found = False

        assert len(x_y) == len(depths)

        return numpy.array(x_y), numpy.array(depths)

    def plot_bezier(
        self,
        as_lc=None,
        pad=5,
        square=None,
        bezier_weight=None,
        segs=None,
        keep_ends=True,
        ax=None,
        **kwargs,
    ):

        if bezier_weight is None:
            bezier_weight = bezier.circular_weight(angle=self.da)

        if segs is None:
            # fastest renderer is the mpl path collection implementation. Try this first if the
            # user hasn't indicated the number of segments in `segs`
            paths = viz.construct_bezier_path_collection(
                self.coords, weight=bezier_weight, keep_ends=keep_ends, **kwargs
            )
            axes = viz.plot_collection(paths, ax=ax)

        else:
            # if the user has asked for `segs`, then compute the bezier using
            # the expensive algorithm.
            _, _ = self._compute_bezier(
                bezier_weight=bezier_weight, segs=segs, keep_ends=keep_ends
            )

            if as_lc or "cmap" in kwargs:

                # if user wishes to render as lc use this route.
                lc = viz.construct_line_collection(self._bezier_coords, **kwargs)
                axes = viz.plot_collection(lc, ax=ax)
            else:
                # if not, plot bezier
                axes = viz.plot(self._bezier_x, self._bezier_y, ax=ax, **kwargs)

        if (ax is None) or square:
            axes = viz.pretty_format_ax(
                axes, coords=self.coords, pad=pad, square=square
            )
        return axes

    def _plot(
        self,
        as_lc=False,
        pad=5,
        square=None,
        ax=None,
        **kwargs,
    ):
        if as_lc or "cmap" in kwargs:
            lc = viz.construct_line_collection(self.coords, **kwargs)
            axes = viz.plot_collection(lc, ax=ax)
        else:
            axes = viz.plot(self.x, self.y, ax=ax, **kwargs)

        if (ax is None) or square:
            axes = viz.pretty_format_ax(
                axes, coords=self.coords, pad=pad, square=square
            )

        return axes

    def plot(
        self,
        *,
        ax=None,
        as_lc=False,
        pad=5,
        square=None,
        as_bezier=False,
        bezier_weight=None,
        segs=None,
        keep_ends=True,
        **kwargs,
    ):

        if any([as_bezier, bezier_weight, segs]):
            return self.plot_bezier(
                ax=ax,
                as_lc=as_lc,
                pad=pad,
                square=square,
                bezier_weight=bezier_weight,
                segs=segs,
                keep_ends=keep_ends,
                **kwargs,
            )

        return self._plot(
            ax=ax,
            as_lc=as_lc,
            pad=pad,
            square=square,
            **kwargs,
        )



import re
from pathlib import Path
from setuptools import setup, find_packages


def search(substr: str, content: str):
    found = re.search(substr, content)
    if found:
        return found.group(1)
    return ""


with open("lsys/__init__.py", encoding="utf8") as f:
    content = f.read()
    version = search(r'__version__ = "(.*?)"', content)
    author = search(r'__author__ = "(.*?)"', content)
    author_email = search(r'__email__ = "(.*?)"', content)


requirements = ["numpy", "matplotlib"]

test_requirements = ["pytest>=3.1", "pytest-mpl>=0.8"]

setup(
    name="lsys",
    version=version,
    description="Create and visualize Lindenmayer systems",
    long_description=Path("./README.md").read_text(),
    long_description_content_type="text/markdown",
    author=author,
    author_email=author_email,
    url="https://github.com/austinorr/lsys",
    packages=find_packages(),
    include_package_data=False,
    install_requires=requirements,
    license="BSD license",
    zip_safe=False,
    keywords=["l-systems", "lindenmayer", "fractal"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
    ],
    python_requires=">=3.8",
    test_suite="lsys.tests",
    tests_require=test_requirements,
    entry_points={
        "console_scripts": [
            "lsys = lsys.cli:main",
        ],
    },
)
